--              Syntax Color
-- These routines are used to display Euphoria program lines
-- in multiple colors. The editor (ed.ex) and the pretty printer (eprint.ex)
-- both include this file.

include graphics.e

-- the built-in keywords
global constant keywords = {
    "and", 
    "by",
    "constant",
    "do",
    "else",
    "elsif",
    "end",
    "exit",
    "for",
    "function",
    "global",
    "if",
    "include",
    "not", 
    "or",
    "procedure", 
    "return",
    "then", 
    "to", 
    "type", 
    "while", 
    "with", 
    "without" }

global sequence
    builtins, library                   -- filled in by create_help

    builtins = {}
    library = {}

global constant NORMAL_COLOR = BRIGHT_WHITE,
                COMMENT_COLOR = WHITE,
                KEYWORD_COLOR = BRIGHT_GREEN,
                BUILTIN_COLOR = YELLOW,
                LIBRARY_COLOR = BRIGHT_MAGENTA,
                STRING_COLOR = BRIGHT_CYAN,
                BRACKET_COLOR = {NORMAL_COLOR, YELLOW, BRIGHT_WHITE, 
                BRIGHT_BLUE, CYAN, GREEN}

-- character classes
constant    DIGIT = 1,
            OTHER = 2,
            LETTER  = 3,
            BRACKET = 4,
            QUOTE   = 5,
            DASH = 6,
            WHITE_SPACE = 7,
            NEW_LINE = 8

global sequence char_class

-- set up character classes for easier line scanning
-- (assume no 0 char)

        char_class = repeat(OTHER, 255)

        char_class['a'..'z'] = LETTER
        char_class['A'..'Z'] = LETTER
        char_class['_'] = LETTER
        char_class['0'..'9'] = DIGIT
        char_class['['] = BRACKET
        char_class[']'] = BRACKET
        char_class['('] = BRACKET
        char_class[')'] = BRACKET
        char_class['{'] = BRACKET
        char_class['}'] = BRACKET
        char_class['\''] = QUOTE
        char_class['"'] = QUOTE
        char_class[' '] = WHITE_SPACE
        char_class['\t'] = WHITE_SPACE
        char_class['\n'] = NEW_LINE
        char_class['-'] = DASH


sequence segment, segments
sequence line               -- the line being processed
integer segStart, segEnd    -- start and end of current segment of line
integer color               -- the current color

procedure flushSeg(integer newColor)

        -- if the color is changing, start a new segment

        -- color changed?
        if newColor != color then
                -- build segment, and append it
                segment = line[segStart..segEnd]
                segments = append( segments, { color, segment }  )
                segment = ""
                segStart = segEnd+1
                color = newColor
        end if

end procedure

global function BuildColorLine(sequence pline)

        -- color code a line of text. the text is broken up into segments in the
        -- form { colorPair, text }

        integer class, last, i, c, bracketLevel
        sequence word

        line = pline
        color = WHITE -- default color
        bracketLevel = 0
        segStart = 1
        segEnd = 0
        segments = {}
        segment = ""

        while 1 do

        -- past end?
        if segEnd = length( line ) then
                exit
        end if

        -- get the character            
        c = line[segEnd+1]

        -- get the class
        if c then
            -- find value
            class = char_class[c]
        else
            -- null
            class = OTHER
        end if

        -- act on class of character

        if class = WHITE_SPACE then
                
                -- continue
                segEnd = segEnd + 1

        elsif class = LETTER then
                last = length(line)
                for j = segEnd + 2 to last do
                        c = line[j]
                        class = char_class[c]
                        if class != LETTER then
                                if class != DIGIT then
                                        last = j - 1
                                        exit
                                end if
                        end if
                end for

                word = line[segEnd+1..last]

                if find(word, keywords) then
                        flushSeg(KEYWORD_COLOR)
                
                elsif find(word, builtins) then
                        flushSeg(BUILTIN_COLOR)
                        
                elsif find(word, library) then
                        flushSeg(LIBRARY_COLOR)
                        
                else
                        flushSeg(NORMAL_COLOR)
                end if
                
                segEnd = last

        elsif class <= OTHER then -- DIGIT too
                flushSeg(NORMAL_COLOR)
                segEnd = segEnd + 1

                elsif class = BRACKET then

                        if find(c, "([{") then
                                bracketLevel = bracketLevel + 1
                        end if

                        if  bracketLevel >= 1
                        and bracketLevel <= length(BRACKET_COLOR) then
                                flushSeg(BRACKET_COLOR[bracketLevel])
                        else
                                flushSeg(NORMAL_COLOR)
                        end if

                        if find(c, ")]}") then
                                bracketLevel = bracketLevel - 1
                        end if

                        segEnd = segEnd + 1

                elsif class = DASH then

                        -- next char also a dash?   
                        if length( line ) > segEnd+1 then
                            if line[segEnd+2] = '-' then
                                    flushSeg(COMMENT_COLOR)
                                    segEnd = length(line)-1
                                    exit
                            end if
                        end if

                        flushSeg(NORMAL_COLOR)
                        segEnd = segEnd + 1

                else  -- QUOTE
                        i = segEnd + 2
                        while i < length(line) do

                                -- end quote of same type?
                                if line[i] = c then
                                        -- move past & exit
                                        i = i + 1
                                        exit

                                elsif line[i] = '\\' then
                                        -- skip; it's an embedded code
                                        if i < length(line)-1 then
                                                i = i + 1 -- ignore escaped char
                                        end if
                                end if


                                i = i + 1

                        end while

                        -- output the string
                        flushSeg(STRING_COLOR)
                        segEnd = i - 1

                end if

        end while

        -- build final segment
        segment = line[segStart..segEnd]
        if length( segment ) != 0 then
                segments = append( segments, { color, segment }  )
        end if

        return segments

end function

