-- dot_pp.e v.1.1
-- pre-processor that adds dot notation to Euphoria
-- extended to match with ee
--
-- (c) 1998 David Cuny
-- email: dcuny@lanset.com
--

-- replacement pp file
global constant PP_LOADED = 1


include file.e
include get.e
include wildcard.e

without warning

sequence inName, outName, errText
integer outFile

errText = ""

sequence 
    includedFile,       -- files already included
    globalFunc,         -- global function
    globalProc,         -- global procedure
    localFunc,          -- local function
    localProc           -- local procedure
    
    includedFile = {}                        
    
    globalProc = {
        "puts", "position", 
        "print", "printf", 
        "clear_screen",    
        "close", "trace",
        "system", 
        "machine_proc", "abort", "poke", 
        "call", 
        "mem_copy", "mem_set",
        "c_proc", "call_proc", 
        "poke4" }

    globalFunc = {
        "length", "integer", "sequence", "object",
        "append", "prepend", "floor", "getc", "gets", "get_key",
        "rand", "repeat", "atom", "compare", "find", "match",
        "time", "command_line", "open", "getenv",
        "sqrt", "sin", "cos", "tan", "log", "date", "remainder",
        "power", "machine_func", "peek", 
        "sprintf", "arctan", "and_bits", "or_bits", "xor_bits",
        "not_bits", "get_pixel", 
        "c_func", "routine_id", "call_func", 
        "peek4s", "peek4u", "&" } 
        
    localProc = {}
    localFunc = {}
    

-----------------------------------------------------------------------------
function butLast( sequence s )
    -- return everything but last item
    return s[1..length(s)-1]
end function

-----------------------------------------------------------------------------
procedure errMsg( sequence error, sequence file, sequence errLine, 
                    integer lineNum, integer linePos )
                    
    -- generate an error in ex.err format
    integer errFile
    
    errFile = open( "ex.err", "w" )

    -- if primary input file, change to output file
    if compare( file, inName ) = 0 then
        file = outName
    end if
    
    -- file:line number
    printf( errFile, "%s:%d\n", {file,lineNum})
    
    -- error message
    printf( errFile, "%s\n", {error} )
    
    -- error line - line feed not needed
    printf( errFile, "%s", {errLine} )
    
    -- pointer at position
    if linePos > 1 then
        puts( errFile, repeat( ' ', linePos ) )
    end if
    puts( errFile, "^\n")
    
    -- close file
    close( errFile )
    
end procedure


-----------------------------------------------------------------------------
constant
    Space = " \t",
    WhiteSpace = Space & "\n"
-----------------------------------------------------------------------------
function getIndent( sequence s )
    -- return leading spaces
    for i = 1 to length( s ) do
        -- non blank?
        if not find( s[i], Space ) then
            return s[1..i-1] 
        end if
    end for
    return  s
end function

                                            
constant
    Delimiter = WhiteSpace & "\n.,{}()[]=<>!+-*/&'\""
-----------------------------------------------------------------------------
function parse( sequence s )
    -- parse into delimited words
    sequence indent, words, padding
    integer this
    
    -- get indent
    indent = getIndent( s )
    
    -- get what's left over
    s = s[length(indent)+1..length(s)]

    -- init word list
    words = { {} }
    padding = { {} }

    for i = 1 to length(s) do
        if find( s[i], Delimiter ) then
        
            -- white space?
            if find( s[i], WhiteSpace ) then
                -- get last word
                this = length( padding )
                padding[this] = padding[this] & s[i]

            else
                -- delimiter is new word
                words = append( words, {s[i]} )
                padding = append( padding, {} )

            end if
                        
        else  
            -- was there a prior character?
            if i > 2 then
                if find( s[i-1], Delimiter ) then 
                    -- beginning of new word
                    words = append( words, {} )
                    padding = append( padding, {} )
                end if
            end if  
            
            -- add to current word
            this = length( words )
            words[this] = words[this] & s[i]

                            
        end if

    end for

    return {words,padding,indent}
    
end function


-----------------------------------------------------------------------------
function getFileName()
    -- get the file name from the command line
    -- if none is found, ask the user for it
    
    sequence args, fName

    -- get the command line
    args = command_line()
    if length( args ) > 2 then
        return args[3]
    end if
    
    -- get prompt from user
    puts( 1, "File to link? " )
    fName = gets(0)
    puts( 1, "\n" )
    
    -- remove line feed
    fName = butLast( fName )
    
    -- empty?
    if length( fName ) = 0 then
        abort(0)
    end if   

    return fName
             
end function            


constant
    EuPlace = getenv( "EUDIR" ),
    Place = { "", EuPlace & "\\", EuPlace & "\\INCLUDE\\" }
-----------------------------------------------------------------------------
function findFile( sequence fName )

    -- returns where a file is
    -- looks in the usual places
    
    -- look in the usual places
    for i = 1 to length( Place ) do
        if sequence( dir( Place[i] & fName ) ) then
            return Place[i] & fName
        end if
    end for
    
    printf( 1, "Unable to locate file %s.\n", {fName} )
    abort(0)
    
end function


-----------------------------------------------------------------------------
function matchWords( sequence s, sequence pattern )

    -- return true if sequence matches pattern
    -- "*" is a wildcard
    
    if length( s ) < length( pattern ) then
        return 0
    end if
    
    for i = 1 to length( pattern ) do         
        -- not wild card?
        if compare( pattern[i], "*" ) then
            -- no match?
            if compare( pattern[i], s[i] )  then
                return 0
            end if
        end if
    end for
    
    -- match
    return 1
    
end function


-----------------------------------------------------------------------------
function getArgName( sequence words, integer at )

    -- extract name from sequence, return position as well
    -- include parenthesis
    integer parens
    sequence word

    parens = 0    
    word = ""
    for i = at to 1 by -1 do
        
        -- add a word
        word = words[i] & word
        
        -- paren check
        if find( words[i], {"]", ")", "}"} ) then
            parens = parens + 1
        end if

        -- can leave if no nesting to worry about        
        if parens = 0 then
            return { word, i }
        end if

        -- this is here so  foo[x]  will pick up 'foo' on the next pass        
        if find( words[i], {"[", "(", "{"} ) then
            parens = parens - 1
        end if

    end for
    
    return { word, 1 }
            
end function


-----------------------------------------------------------------------------
procedure includeFile( sequence fName, integer deep )

    -- open an include file for processing
    integer inFile, argAt, funcFlag, procFlag, lineCount
    sequence words, padding, indent, locals, func, arg, tmp
    object text, data
    
    -- get the file
    fName = findFile( fName )
    
    -- open if
    inFile = open( fName, "r" )
    if inFile = -1 then
        printf( 1, "Error opening file %s.\n", {fName} )
        abort(0)
    end if

    -- save locals from calling file
    locals = { localFunc, localProc }
    localFunc = {}
    localProc = {}

    -- line count
    lineCount = 0
    
    -- loop while there is no error
    while length( errText ) = 0 do
    
        -- get text
        data = gets( inFile )
        
        -- end of file?
        if integer( data ) then
            exit
        end if

        -- line count, for error
        lineCount = lineCount + 1
        
        -- look for include
        text = parse(data) 
        words   = text[1]   -- word list
        padding = text[2]   -- spaces between words
        indent  = text[3]   -- padding

        if matchWords( words, {"global", "function", "*"} ) then
            -- add to global functions
            globalFunc = append( globalFunc, words[3] )
        
        elsif matchWords( words, {"global", "procedure", "*"} ) then
            -- add to global procedures
            globalProc = append( globalProc, words[3] )

        elsif matchWords( words, {"function", "*"} ) then
            -- add to local functions
            localFunc = append( localFunc, words[2] )
        
        elsif matchWords( words, {"procedure", "*"} ) then
            -- add to local procedures
            localProc = append( localProc, words[2] )

        -- include file?            
        elsif matchWords( words, {"include"} ) then

            -- rebuild file name        
            text = ""
            for i = 2 to length( words ) do
                if compare( words[i], "-" ) = 0 then
                    -- comment
                    exit
                else
                    text = text & words[i]
                end if
            end for                    

            -- load include file?
            if not find( text, includedFile ) then
            
                -- add to included list
                includedFile = append( includedFile, text )
                
                -- include file
                includeFile( text, deep + 1 ) 
                
            end if
            
        end if

        -- only substitute for current code
        if deep = 0 then
        
            -- substitution?
            for i = 2 to length( words )-2 do     

                -- form is "arg.func("
                -- not     "..func("
                -- or       "func..("         
                
                if  compare( words[i-1], "." ) != 0  
                and compare( words[i+0], "." ) = 0
                and compare( words[i+1], "." ) != 0
                and compare( words[i+2], "(" ) = 0 then

                    -- clear flags
                    funcFlag = 0
                    procFlag = 0

                    -- find it?                
                    if find( words[i+1], localFunc ) then
                        funcFlag = 1
                        
                    elsif find( words[i+1], globalFunc ) then
                        funcFlag = 1
                        
                    elsif find( words[i+1], localProc ) then
                        procFlag = 1
                        
                    elsif find( words[i+1], globalProc ) then
                        procFlag = 1

                    end if                        

                    -- matched?                        
                    if funcFlag
                    or procFlag then
                        
                        -- replace with a function call
                        func    = words[i+1]
                        tmp     = getArgName( words, i-1 )
                        arg     = tmp[1]   -- arg name
                        argAt   = tmp[2]   -- position
                        
                        -- clear prior text                        
                        for j = argAt to i+2 do
                            words[j] = ""
                        end for

                        -- is this a self assignment?
                        if argAt = 1      
                        and funcFlag then
                            -- assign to self
                            words[argAt] = sprintf( "%s = %s( %s", {arg,func,arg})
                        else
                            -- normal call
                            words[argAt] = sprintf( "%s( %s", {func,arg})
                        end if
                        
                        -- more than one arg?                        
                        if compare( words[i+3], ")" ) then
                            words[argAt] = words[argAt] & ", "
                        else
                              words[argAt] = words[argAt] & " "
                        end if

                    else
                        -- give error message and exit loop
                        tmp = words[i+1]
                        errText = "Unknown routine '" & tmp & "'."
                        errMsg( errText, fName, data, lineCount, 
                                    match( "." & tmp & "(", data ) )
                        
                        -- exit loop
                        
                    end if
                end if
            end for
    
            -- reconstruct                 
            puts( outFile, indent )
            for i = 1 to length( words ) do
                puts( outFile, words[i] & padding[i] )
            end for     
            
        end if
    
    end while                       
    
    close( inFile )

    -- restore locals from calling file
    localFunc = locals[1]
    localProc = locals[2]
    
end procedure

-----------------------------------------------------------------------------
global function preprocess( sequence inFileName, integer verbose )

    -- set to global
    inName = upper(inFileName)

    -- build correct extention
    if match( ".PP", inName ) then
        outName = inName[1..length(inName)-2] & "EX" 
    elsif match( ".P", inName ) then
        outName = inName[1..length(inName)-1] & "E" 
    else
        -- bail
        outName = "OUT.EX"
    end if
    
    -- open outfile
    outFile = open( outName, "w" )

    -- process it    
    includeFile( inName, 0 )

    -- close it
    close( outFile )
    

    -- return error flag and file name
    if length( errText ) = 0 then
        -- no error
        return { 1, outName }
    else
        -- error
        return { 0, outName }
    end if
    
end function

